import json
import os
from collections.abc import Iterator
from dataclasses import dataclass
from subprocess import CalledProcessError
from typing import Any, Optional

import httpx
from pydantic import Field

from ..driver import OS, Architecture
from ..filters import Filter, Filters, choose
from ..utils import (
    allow_singular_none,
    async_cached,
    async_cached_lock,
    http_get,
    run_output,
)
from .archive import ArchivePackage


@async_cached_lock
async def github_auth_token() -> Optional[str]:
    try:
        return os.environ["GITHUB_TOKEN"]
    except KeyError:
        pass

    try:
        return await run_output("gh", "auth", "token", silent=True)
    except (CalledProcessError, FileNotFoundError):
        pass

    return None  # pragma: no cover


async def github_api(url: str) -> Any:
    token = await github_auth_token()

    headers = {}
    if token:
        headers["Authorization"] = f"token {token}"

    result = await http_get(f"https://api.github.com/{url}", headers=headers)

    return json.loads(result)


@dataclass
class GitHubReleaseArtifact:
    name: str
    url: str


@dataclass
class GitHubRelease:
    id: int
    tag_name: str
    prerelease: bool
    assets: list[GitHubReleaseArtifact]

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> "GitHubRelease":
        return cls(
            id=data["id"],
            tag_name=data["tag_name"],
            prerelease=data["prerelease"],
            assets=[
                GitHubReleaseArtifact(
                    name=result["name"], url=result["browser_download_url"]
                )
                for result in data["assets"]
            ],
        )


ARCHITECTURE_FILTERS: dict[str, list[str]] = {
    "aarch64": ["arm64", "arm"],
    "i386": ["i686", "x86"],
    "mips": [],
    "powerpc": ["ppc"],
    "s390x": [],
    "x86_64": ["amd64", "x64"],
}

OS_FILTERS: dict[str, list[str]] = {
    "darwin": ["macOS", "macos", "osx"],
    "linux": [],
    "windows": [],
}


class GitHubPackage(ArchivePackage, Filters):
    repo: str
    skip_releases: list[str] = Field(default_factory=list, alias="skip_release")
    skip_releases_val = allow_singular_none("skip_releases")

    @async_cached
    async def releases(self) -> list[GitHubRelease]:
        result = await github_api(f"repos/{self.repo}/releases")
        return [GitHubRelease.from_json(release) for release in result]

    @async_cached
    async def latest_release(self) -> Optional[GitHubRelease]:
        try:
            result = await github_api(f"repos/{self.repo}/releases/latest")
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            raise  # pragma: no cover
        return GitHubRelease.from_json(result)

    def want_release(self, release: GitHubRelease) -> bool:
        if release.prerelease:
            return False
        if release.tag_name in self.skip_releases:
            return False
        return True

    async def release(self) -> GitHubRelease:
        latest = await self.latest_release()
        if latest and self.want_release(latest):
            return latest

        candidates = await self.releases()
        for candidate in candidates:
            if not self.want_release(candidate):
                continue
            return candidate
        raise ValueError(f"No releases found for {self.repo}.")

    @classmethod
    def environment_filters(
        cls, *, target_os: OS, target_arch: Architecture
    ) -> Iterator[Filter[str]]:
        for signature_hint in [".asc", ".sig", "sha256", "sha512", ".yml"]:
            yield cls.excludes_(signature_hint)

        for system_package_hint in [".deb", ".rpm", ".dmg", ".exe", ".appimage"]:
            yield cls.excludes_(system_package_hint)

        yield from cls.from_synonyms(
            OS_FILTERS, target_os.switch(linux="linux", macos="darwin")
        )

        yield from cls.from_synonyms(ARCHITECTURE_FILTERS, target_arch)

        if target_os.switch(linux=True, macos=False):
            yield cls.includes_("gnu")
            yield cls.excludes_("musl")  # pragma: no cover

    def all_filters(
        self, *, target_os: OS, target_arch: Architecture
    ) -> Iterator[Filter[str]]:
        yield from self.filters()
        yield from self.environment_filters(
            target_os=target_os, target_arch=target_arch
        )

    async def artifact(self) -> GitHubReleaseArtifact:
        candidates = (await self.release()).assets

        def candidate_filter(
            name_filter: Filter,
        ) -> Filter[GitHubReleaseArtifact]:
            return lambda candidate: name_filter(candidate.name)

        target_os = await self.driver.os()
        target_arch = await self.driver.architecture()
        return choose(
            candidates,
            map(
                candidate_filter,
                self.all_filters(target_os=target_os, target_arch=target_arch),
            ),
        )

    async def archive_url(self) -> str:
        return (await self.artifact()).url

    def derive_name(self) -> str:
        return self.repo

    async def get_remote_version(self) -> str:
        release = await self.release()
        return str(release.id)



from functools import cached_property
from typing import Optional

import trio

from ..tracker import Tracker
from .installer.flatpak import Flatpak
from .manual_version import ManualVersion

FLATPAK_LOCK = trio.Lock()


class FlatpakPackage(ManualVersion):
    flatpak: str

    def derive_name(self) -> str:
        return self.flatpak

    @cached_property
    def installer(self) -> Flatpak:
        return Flatpak(self.driver)

    async def get_remote_version(self) -> str:
        return await self.installer.latest_version(self.flatpak)

    async def local_version(self) -> Optional[str]:
        return await self.installer.installed_version(self.flatpak)

    async def install(self, *, tracker: Tracker) -> None:
        async with FLATPAK_LOCK:
            if await self.local_version():
                # Cannot install an old version of a package in tests
                await self.installer.upgrade(self.flatpak)  # pragma: no cover
            else:
                await self.installer.install(self.flatpak)

        await super().install(tracker=tracker)



from collections.abc import AsyncIterable, Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from .config import MatchConfig, parse_config
from .driver import Driver
from .package import Package, parse_packages
from .parallel import PartialException, PartialResults, parallel_map_progress
from .state import DB, VERSIONS
from .tracker import ManagerTracker, Tracker
from .utils import flatten


@dataclass
class InstallResult:
    installed: list[Package]
    failed: list[tuple[Package, BaseException]]


@dataclass
class Manager:
    db: DB
    driver: Driver
    data_path: Path

    def load_data(self, *path: str) -> Any:
        with open(self.data_path / Path(*path)) as f:
            return yaml.safe_load(f)

    def load_config(self) -> list[MatchConfig]:
        return parse_config(self.load_data("mybox.yaml"), driver=self.driver)

    def load_component(self, component: str) -> Sequence[Package]:
        packages = self.load_data("packages", f"{component}.yaml")
        return parse_packages(packages, db=self.db, driver=self.driver)

    def load_components(self, components: frozenset[str]) -> list[Package]:
        return flatten(self.load_component(component) for component in components)

    async def cleanup(self, packages: list[Package]) -> None:
        package_names = set(package.name for package in packages)

        versions = VERSIONS(self.db)

        all_versions = list(versions.find_ids())

        for package, _ in all_versions:
            if package not in package_names:
                versions.delete(id=package)

    async def install(self) -> InstallResult:
        config = self.load_config()
        components = await MatchConfig.components(config)
        packages = self.load_components(frozenset(components))

        return await self.install_packages(packages)

    async def install_packages(self, packages: list[Package]) -> InstallResult:
        async with Tracker.tracking(driver=self.driver, db=self.db) as tracker:

            async def process_and_record(package: Package) -> Iterable[Package]:
                return [pkg async for pkg in self.install_package(tracker, package)]

            try:
                results = await parallel_map_progress(process_and_record, packages)

                await self.cleanup(packages)

                return InstallResult(installed=flatten(results), failed=[])
            except PartialResults as e:
                installed: list[Package] = []
                failed: list[tuple[Package, BaseException]] = []

                for package, result in zip(packages, e.results):
                    if isinstance(result, PartialException):
                        failed.append((package, result.exception))
                    else:
                        installed.extend(result.result)

                return InstallResult(installed=installed, failed=failed)

    async def install_package(
        self, tracker: ManagerTracker, package: Package
    ) -> AsyncIterable[Package]:
        async for prerequisite in package.prerequisites():
            async for result in self.install_package(tracker, prerequisite):
                yield result

        try:
            if not await package.applicable():
                return

            if await package.is_installed():
                tracker.skip(package.name)
                return

            await package.install(tracker=tracker.track(package.name))
        except:
            tracker.skip(package.name)
            raise

        yield package



import subprocess
from collections.abc import Awaitable, Callable, Iterable
from functools import wraps
from importlib.metadata import version as importlib_version
from pathlib import Path
from typing import Any, Optional, TypeVar, overload

import httpx
import trio
from pydantic import field_validator

MYBOX_VERSION = importlib_version("mybox")

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


Some = Optional[T | list[T]]


def unsome_(x: Some[T]) -> Optional[list[T]]:
    if x is None:
        return None
    if isinstance(x, list):
        return x
    return [x]


def unsome(x: Some[T]) -> list[T]:
    return unsome_(x) or []


def allow_singular(field: str) -> Any:
    return field_validator(field, mode="before")(lambda cls, x: unsome_(x))


def allow_singular_none(field: str) -> Any:
    return field_validator(field, mode="before")(lambda cls, x: unsome(x))


def matches_if_specified(possibilities: Optional[list[T]], value: T) -> bool:
    return possibilities is None or value in possibilities


def intercalate(delimiter: T, items: Iterable[Iterable[T]]) -> list[T]:
    it = iter(items)
    try:
        result = list(next(it))
    except StopIteration:
        return []
    for x in it:
        result.append(delimiter)
        result.extend(list(x))
    return result


RunArg = str | Path


async def run(*args: RunArg) -> subprocess.CompletedProcess:
    return await trio.run_process(args, check=True)


async def run_ok(*args: RunArg) -> bool:
    try:
        result = await trio.run_process(
            args, check=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        return result.returncode == 0
    except FileNotFoundError:
        return False


async def run_output(*args: RunArg, silent: bool = False) -> str:
    result = await trio.run_process(
        args, capture_stdout=True, capture_stderr=silent, check=True
    )
    return result.stdout.decode().strip()


def flatten(items: Iterable[Iterable[T]]) -> list[T]:
    return [item for sublist in items for item in sublist]


http_client = httpx.AsyncClient(
    headers={"user-agent": f"mybox/{MYBOX_VERSION} (github.com/koterpillar/mybox)"}
)


async def http_get(url: str, headers: Optional[dict[str, str]] = None) -> str:
    response = await http_client.get(url, headers=headers)
    response.raise_for_status()
    return response.text


async def url_version(url: str) -> str:
    head_response = await http_client.head(url, follow_redirects=True)
    return head_response.headers["etag"]


GIT_PREFIX = "git+"


async def repo_version(repo: str, *, branch: Optional[str] = None) -> str:
    if repo.startswith(GIT_PREFIX):
        repo = repo.removeprefix(GIT_PREFIX)
    if not branch:
        branch = "HEAD"
    return (await run_output("git", "ls-remote", repo, branch)).split()[0]


def with_extensions(name: str, extensions: list[str]) -> list[str]:
    if any(name.endswith(f".{ext}") for ext in extensions):
        return [name]
    return [f"{name}.{ext}" for ext in extensions]


def raise_(exception: BaseException) -> Any:
    raise exception


@overload
def async_cached(fn: Callable[[], Awaitable[T]]) -> Callable[[], Awaitable[T]]: ...


@overload
def async_cached(fn: Callable[[U], Awaitable[T]]) -> Callable[[U], Awaitable[T]]: ...


@overload
def async_cached(
    fn: Callable[[U, V], Awaitable[T]],
) -> Callable[[U, V], Awaitable[T]]: ...


def async_cached(fn: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    return _async_cached_lock(None, fn)


@overload
def async_cached_lock(fn: Callable[[], Awaitable[T]]) -> Callable[[], Awaitable[T]]: ...


@overload
def async_cached_lock(
    fn: Callable[[U], Awaitable[T]],
) -> Callable[[U], Awaitable[T]]: ...


@overload
def async_cached_lock(
    fn: Callable[[U, V], Awaitable[T]],
) -> Callable[[U, V], Awaitable[T]]: ...


def async_cached_lock(fn: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    return _async_cached_lock(trio.Lock(), fn)


class NoLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args: Any) -> None:
        pass


NO_LOCK = NoLock()


def _async_cached_lock(
    lock: Optional[trio.Lock], fn: Callable[..., Awaitable[T]]
) -> Callable[..., Awaitable[T]]:
    cache: dict[Any, T] = {}

    @wraps(fn)
    async def cached_fn(*args: Any) -> T:
        if args not in cache:
            async with lock or NO_LOCK:
                # Check again, someone holding the lock might have updated the cache
                if args not in cache:
                    cache[args] = await fn(*args)
        return cache[args]

    return cached_fn



from abc import ABC, abstractmethod
from collections.abc import AsyncIterable, Iterable
from pathlib import Path

from pydantic import Field

from ..extractor import get_extractor, get_single_extractor
from ..tracker import Tracker
from ..utils import allow_singular_none, async_cached, with_extensions
from .base import Package
from .manual import ManualPackage
from .system import SystemPackage

ICON_EXTENSIONS = ["svg", "png"]

FONT_EXTENSIONS = ["ttf", "otf"]


class ArchivePackage(ManualPackage, ABC):
    raw: bool | str = False

    binary_paths: list[str] = Field(default_factory=list, alias="binary_path")
    binary_paths_val = allow_singular_none("binary_paths")

    @abstractmethod
    async def archive_url(self) -> str:
        pass

    @property
    def pathname(self) -> str:
        return self.name.replace("/", "--")

    @async_cached
    async def package_directory(self) -> Path:
        return (await self.driver.local()) / "mybox" / self.pathname

    async def extract(self, url: str, source: Path) -> None:
        if self.raw:
            if isinstance(self.raw, str):
                filename = self.raw
            else:
                filename = url.rsplit("/", 1)[-1]
            await self.extract_raw(url, source, filename)

        else:
            extractor = await get_extractor(url, driver=self.driver)
            await extractor.extract(
                archive=source, target_directory=await self.package_directory()
            )

    async def extract_raw(self, url: str, source: Path, filename: str) -> None:
        target = await self.package_directory() / filename

        extractor = await get_single_extractor(url, driver=self.driver)
        await extractor.extract(archive=source, target=target)
        if filename in self.binaries:
            await self.driver.make_executable(target)

    async def find_in_package_directory(
        self,
        paths: list[list[str]],
        name: str,
        *,
        require_executable: bool = False,
        target_desc: str,
    ) -> Path:
        for relative_path in paths:
            candidate = await self.package_directory() / Path(*relative_path) / name
            if not await self.driver.is_file(candidate):
                continue
            if require_executable and not await self.driver.is_executable(candidate):
                continue
            return candidate
        raise ValueError(
            f"Cannot find {target_desc} '{name}' in {await self.package_directory()}."
        )

    async def binary_path(self, binary: str) -> Path:
        return await self.find_in_package_directory(
            paths=[[], ["bin"], *([bp] for bp in self.binary_paths)],
            name=binary,
            require_executable=True,
            target_desc="binary",
        )

    async def freedesktop_app_path(self, name: str) -> Path:
        return await self.find_in_package_directory(
            paths=[["share", "applications"]],
            name=f"{name}.desktop",
            target_desc="application",
        )

    async def icon_paths(self, name: str) -> Iterable[Path]:
        names = with_extensions(name, ICON_EXTENSIONS)
        return await self.driver.find(await self.package_directory(), name=names)

    async def font_path(self, name: str) -> Path:
        names = with_extensions(name, FONT_EXTENSIONS)
        candidates = await self.driver.find(await self.package_directory(), name=names)
        if candidates:
            return candidates[0]
        else:
            raise ValueError(
                f"Cannot find font '{name}' in {await self.package_directory()}."
            )

    async def install(self, *, tracker: Tracker):
        url = await self.archive_url()
        async with self.driver.tempfile() as archive_path:
            await self.driver.run(
                "curl",
                "--silent",
                "--show-error",
                "--location",
                "--output",
                archive_path,
                url,
            )

            package_directory = await self.package_directory()
            await self.driver.makedirs(package_directory)
            tracker.track(package_directory.parent, root=self.root)
            tracker.track(package_directory, root=self.root)

            await self.extract(url, archive_path)

        await super().install(tracker=tracker)

    async def prerequisites(self) -> AsyncIterable[Package]:
        async for package in super().prerequisites():
            yield package  # pragma: no cover

        os = await self.driver.os()

        for system in os.switch_(
            linux=lambda linux: (
                []
                if linux.distribution == "libreelec"
                else [
                    "bzip2",
                    "unzip",
                    {"debian": "xz-utils", "ubuntu": "xz-utils", "fedora": "xz"}[
                        linux.distribution
                    ],
                ]
            ),
            macos=lambda: ["gnu-tar", "xz"],
        ):
            yield SystemPackage(
                system=system,
                db=self.db,
                driver=self.driver_,
            )



import json
from dataclasses import dataclass
from typing import Optional

from ...driver import RunResultOutput
from ...utils import async_cached, async_cached_lock, http_get
from .base import PackageCacheInstaller, PackageVersionInfo


@dataclass
class BrewPackageVersionInfo:
    """
    Brew installer prefers casks to formulae with the same name. Brew itself
    reports both separately, but it requires specifying homebrew/cask/
    everywhere so it's easier for the users to assume a cask if one exists.
    """

    cask: Optional[PackageVersionInfo]
    formula: Optional[PackageVersionInfo]

    @staticmethod
    def formula_version(version: str) -> str:
        """
        To distinguish a cask from a formula of the same name, especially if
        they have the same versions, prepend the formula version with
        "formula_".
        """
        return f"formula_{version}"

    @property
    def installed(self) -> Optional[str]:
        """
        Report installed state of both cask and formula so that any state but
        "cask is installed and formula isn't" results in a different version.
        """
        cask_installed = self.cask.installed if self.cask else None
        formula_installed = (
            self.formula_version(self.formula.installed)
            if self.formula and self.formula.installed
            else None
        )

        if cask_installed and formula_installed:
            return f"{cask_installed} {formula_installed}"
        return cask_installed or formula_installed

    @property
    def latest(self) -> str:
        if self.cask:
            return self.cask.latest
        if self.formula:
            return self.formula_version(self.formula.latest)
        raise ValueError("No cask or formula info.")  # pragma: no cover


class Brew(PackageCacheInstaller[BrewPackageVersionInfo]):
    async def tap(self, repo: str) -> None:
        await self.driver.run(await self.brew(), "tap", repo)

    async def tapped(self) -> set[str]:
        output = await self.driver.run_output(await self.brew(), "tap")
        return set(output.strip().splitlines())

    async def install(self, package: str) -> None:
        info = await self.package_info(package)
        if info.cask:
            await self.driver.run(await self.brew(), "install", "--cask", package)
        else:
            await self.driver.run(await self.brew(), "install", package)
        await super().install(package)

    async def upgrade(self, package: str) -> None:
        info = await self.package_info(package)
        if not info.installed:
            raise ValueError(
                "Expected package to be installed when upgrading."
            )  # pragma: no cover
        if info.cask:
            if info.formula and info.formula.installed:
                # We want a cask, not a formula, uninstall the formula
                await self.driver.run(await self.brew(), "uninstall", package)
            if info.cask.installed:
                # Both formula and cask were installed; formula was uninstalled,
                # upgrade the cask if needed
                if info.cask.latest != info.cask.installed:
                    # Can't install an old version of a package in tests
                    await self.driver.run(
                        await self.brew(), "upgrade", "--cask", package
                    )  # pragma: no cover
            else:
                # Formula was installed (thus upgrading), but cask wasn't
                await self.driver.run(await self.brew(), "install", "--cask", package)
        else:
            # Can't install an old version of a package in tests
            await self.driver.run(
                await self.brew(), "upgrade", package
            )  # pragma: no cover
        await super().upgrade(package)

    @async_cached
    async def brew(self) -> str:
        # https://docs.brew.sh/FAQ#why-is-the-default-installation-prefix-opthomebrew-on-apple-silicon
        return await self.driver.find_executable("brew", "/opt/homebrew/bin/brew")

    @async_cached_lock
    async def brew_update(self) -> None:
        await self.driver.run(await self.brew(), "update")

    async def brew_info(self, *args: str) -> RunResultOutput:
        return await self.driver.run_output_(
            await self.brew(), "info", "--json=v2", *args
        )

    async def brew_info_cask(self, cask: str) -> RunResultOutput:
        return await self.brew_info(f"homebrew/cask/{cask}")

    async def get_package_info(
        self, package: Optional[str]
    ) -> dict[str, BrewPackageVersionInfo]:
        await self.brew_update()

        brew_result: RunResultOutput
        if not package:
            brew_result = await self.brew_info("--installed")
        elif "/" not in package:
            # When a bare package name is given (without homebrew/, etc. prefix),
            # try to look up the cask first.
            brew_result = await self.brew_info_cask(package)
            if not brew_result.ok:
                brew_result = await self.brew_info(package)
        else:
            brew_result = await self.brew_info(package)

        results: dict[str, BrewPackageVersionInfo] = {}

        if not brew_result.ok:
            return results
        info = json.loads(brew_result.output)

        for cask in info["casks"]:
            if cask["tap"] == "homebrew/cask":
                # Do not require prefixes for casks from homebrew/cask
                name = cask["token"]
            else:
                name = f"{cask['tap']}/{cask['token']}"
            results[name] = BrewPackageVersionInfo(
                cask=PackageVersionInfo(
                    installed=cask["installed"], latest=cask["version"]
                ),
                formula=None,
            )

        for formula in info["formulae"]:
            name = formula["name"]
            try:
                installed = formula["installed"][0]["version"]
            except IndexError:
                installed = None
            latest = self.formula_version(formula)
            pvi = PackageVersionInfo(installed=installed, latest=latest)

            if installed and name in results:
                # Cask with the same name exists. If a _formula_ is installed
                # instead, record the formula version as installed too, so the
                # package will be updated to cask.
                results[name].formula = pvi
            else:
                results[name] = BrewPackageVersionInfo(cask=None, formula=pvi)

        # If queried for installed packages, check for any casks with the same
        # names as installed formulae
        if not package:
            cask_response = await http_get("https://formulae.brew.sh/api/cask.json")

            all_casks = json.loads(cask_response)
            for cask in all_casks:
                name = cask["token"]
                if name not in results:
                    continue  # No formula with this name installed, can ignore
                if results[name].cask:
                    continue  # Already know about this cask
                # Record this cask as available but not installed (it would have
                # been already found otherwise)
                results[name].cask = PackageVersionInfo(
                    installed=None, latest=cask["version"]
                )

        return results

    @staticmethod
    def formula_version(info: dict) -> str:
        version = info["versions"]["stable"]
        revision = info.get("revision")
        if revision:
            version += f"_{revision}"
        return version



from typing import Optional

from ..base import PackageArgs
from .base import PackageCacheInstaller, PackageVersionInfo


class Flatpak(PackageCacheInstaller[PackageVersionInfo]):
    REPO_URL = "https://dl.flathub.org/repo/flathub.flatpakrepo"
    REPO_NAME = "flathub"

    # FIXME: postinstall for system packages should always run as root
    FLATPAK: PackageArgs = {
        "system": "flatpak",
        "os": "linux",
        "service": "dbus",
        "post": f"sudo flatpak remote-add --if-not-exists {REPO_NAME} {REPO_URL}",
    }

    async def install(self, package: str) -> None:
        await self.driver.run("flatpak", "install", "-y", self.REPO_NAME, package)
        await super().install(package)

    # can't install an old version of a package in tests
    async def upgrade(self, package: str) -> None:  # pragma: no cover
        await self.driver.run("flatpak", "upgrade", "-y", package)
        await super().upgrade(package)

    @staticmethod
    def parse_versions(output: str) -> dict[str, str]:
        versions = {}
        for line in output.splitlines():
            name, origin, commit = line.split()
            versions[name] = f"{origin}:{commit}"
        return versions

    async def get_installed(self) -> dict[str, str]:
        result = await self.driver.run_output(
            "flatpak", "list", "--app", "--columns=application,origin,active"
        )
        return self.parse_versions(result)

    async def get_latest(self) -> dict[str, str]:
        result = await self.driver.run_output(
            "flatpak", "remote-ls", "--app", "--columns=application,origin,commit"
        )
        return self.parse_versions(result)

    async def get_package_info(
        self, _package: Optional[str]
    ) -> dict[str, PackageVersionInfo]:
        # FIXME: Cannot selectively query for only a single package

        latest = await self.get_latest()
        installed = await self.get_installed()

        return {
            name: PackageVersionInfo(
                installed=installed.get(name), latest=latest_version
            )
            for name, latest_version in latest.items()
        }



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, Protocol, TypeVar

import trio

from ...driver import Driver


class Installer(ABC):
    def __init__(self, driver: Driver) -> None:
        self.driver = driver
        super().__init__()

    @abstractmethod
    async def install(self, package: str) -> None:
        pass

    @abstractmethod
    async def upgrade(self, package: str) -> None:
        pass

    @abstractmethod
    async def installed_version(self, package: str) -> Optional[str]:
        pass

    @abstractmethod
    async def latest_version(self, package: str) -> str:
        pass


class PVP(Protocol):
    @property
    def installed(self) -> Optional[str]: ...
    @property
    def latest(self) -> str: ...


PV = TypeVar("PV", bound=PVP)


@dataclass
class PackageVersionInfo:
    installed: Optional[str]
    latest: str


class PackageCacheInstaller(Generic[PV], Installer, ABC):
    cache: Optional[dict[str, PV]]

    def __init__(self, driver: Driver) -> None:
        self.cache = None
        self.lock = trio.Lock()
        super().__init__(driver)

    @abstractmethod
    async def get_package_info(self, package: Optional[str]) -> dict[str, PV]:
        pass

    async def package_info(self, package: str) -> PV:
        async with self.lock:
            if self.cache is None:
                self.cache = await self.get_package_info(None)

        async with self.lock:
            if package not in self.cache:
                self.cache.update(await self.get_package_info(package))

        try:
            return self.cache[package]
        except KeyError:
            raise ValueError(f"Unknown package: {package}") from None

    async def installed_version(self, package: str) -> Optional[str]:
        info = await self.package_info(package)
        return info.installed

    async def latest_version(self, package: str) -> str:
        info = await self.package_info(package)
        return info.latest

    async def invalidate_cache(self, package: str) -> None:
        async with self.lock:
            if self.cache:
                self.cache.pop(package, None)

    async def install(self, package: str) -> None:
        await super().install(package)
        await self.invalidate_cache(package)

    async def upgrade(self, package: str) -> None:  # pragma: no cover
        # Can't install an old version of a package in tests
        await super().upgrade(package)
        await self.invalidate_cache(package)
